package http

import (
	"embed"
	"fmt"
	"html/template"
	"io/fs"
	"net/http"
	"net/url"
	"path/filepath"
	"strings"
	"time"

	"github.com/mediocregopher/blog.mediocregopher.com/srv/http/apiutil"
)

//go:embed tpl
var tplFS embed.FS

func mustReadTplFile(fileName string) string {
	path := filepath.Join("tpl", fileName)

	b, err := fs.ReadFile(tplFS, path)
	if err != nil {
		panic(fmt.Errorf("reading file %q from tplFS: %w", path, err))
	}

	return string(b)
}

func (a *api) blogURL(base *url.URL, path string, abs bool) string {
	// filepath.Join strips trailing slash, but we want to keep it
	trailingSlash := strings.HasSuffix(path, "/")

	path = filepath.Join("/", base.Path, path)

	if trailingSlash && path != "/" {
		path += "/"
	}

	if !abs {
		return path
	}

	u := *base
	u.Path = path
	return u.String()
}

func (a *api) postURL(id string, abs bool) string {
	path := filepath.Join("posts", id)
	return a.blogURL(a.params.PublicURL, path, abs)
}

func (a *api) editPostURL(id string, abs bool) string {
	return a.postURL(id, abs) + "?method=edit"
}

func (a *api) managePostsURL(abs bool) string {
	return a.blogURL(a.params.PublicURL, "posts?method=manage", abs)
}

func (a *api) manageAssetsURL(abs bool) string {
	return a.blogURL(a.params.PublicURL, "assets?method=manage", abs)
}

func (a *api) assetURL(id string, abs bool) string {
	path := filepath.Join("assets", id)
	return a.blogURL(a.params.PublicURL, path, false)
}

func (a *api) draftPostURL(id string, abs bool) string {
	path := filepath.Join("drafts", id)
	return a.blogURL(a.params.PublicURL, path, abs)
}

func (a *api) editDraftPostURL(id string, abs bool) string {
	return a.draftPostURL(id, abs) + "?method=edit"
}

func (a *api) manageDraftPostsURL(abs bool) string {
	return a.blogURL(a.params.PublicURL, "drafts", abs) + "?method=manage"
}

func (a *api) draftsURL(abs bool) string {
	return a.blogURL(a.params.PublicURL, "drafts", abs)
}

func (a *api) tplFuncs() template.FuncMap {
	return template.FuncMap{
		"StaticInlineCSS": func(path string) (template.CSS, error) {
			path = filepath.Join("static", path)
			b, err := staticFS.ReadFile(path)
			return template.CSS(b), err
		},
		"DraftURL": func(id string) string {
			return a.draftPostURL(id, false)
		},
		"DateTimeFormat": func(t time.Time) string {
			return t.Format("2006-01-02")
		},
		"SafeURL": func(u string) template.URL {
			return template.URL(u)
		},
	}
}

func (a *api) parseTpl(name, tplBody string) (*template.Template, error) {

	tpl := template.New(name)
	tpl = tpl.Funcs(a.tplFuncs())
	tpl = tpl.Funcs(template.FuncMap(a.postPreprocessFuncs().ToFuncMap()))

	var err error

	if tpl, err = tpl.Parse(tplBody); err != nil {
		return nil, err
	}

	return tpl, nil
}

func (a *api) mustParseTpl(name string) *template.Template {
	return template.Must(a.parseTpl(name, mustReadTplFile(name)))
}

func (a *api) mustParseBasedTpl(name string) *template.Template {
	tpl := a.mustParseTpl(name)
	tpl = template.Must(tpl.New("gemini-cta.html").Parse(mustReadTplFile("gemini-cta.html")))
	tpl = template.Must(tpl.New("base.html").Parse(mustReadTplFile("base.html")))
	return tpl
}

type tplData struct {
	Payload interface{}
}

func newTPLData(r *http.Request, payload interface{}) tplData {
	return tplData{
		Payload: payload,
	}
}

// executeTemplate expects to be the final action in an http.Handler
func executeTemplate(
	rw http.ResponseWriter, r *http.Request,
	tpl *template.Template, payload interface{},
) {

	tplData := newTPLData(r, payload)

	if err := tpl.Execute(rw, tplData); err != nil {
		apiutil.InternalServerError(
			rw, r, fmt.Errorf("rendering template: %w", err),
		)
		return
	}
}

func (a *api) executeRedirectTpl(
	rw http.ResponseWriter, r *http.Request, url string,
) {
	executeTemplate(rw, r, a.redirectTpl, struct {
		URL string
	}{
		URL: url,
	})
}

func (a *api) renderDumbTplHandler(tplName string) http.Handler {

	tpl := a.mustParseBasedTpl(tplName)

	return http.HandlerFunc(func(rw http.ResponseWriter, r *http.Request) {

		tplData := newTPLData(r, nil)

		if err := tpl.Execute(rw, tplData); err != nil {
			apiutil.InternalServerError(
				rw, r, fmt.Errorf("rendering %q: %w", tplName, err),
			)
			return
		}
	})
}
